/*
 *-----------------------------------------------------------------
 *
 * Name: chassisd-bc.c
 *
 * Copyright (C) 2004, Onstor, Inc.
 *
 * Description:
 *
 *   Chassis manager daemon code for Bobcat.
 *
 * Created by:  Tony Wong
 *
 * Date Created:    7/01/04
 *
 *-----------------------------------------------------------------
 */
#ifdef SSC

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include "../sm-anpssc/anpssc-api.h"
#include "nfx-incl.h"
#include "../sm-elog/elog-api.h"
#include "../sm-event/event-api.h"
#include "../sm-seep/seep.h"
#include "cm.h"
#include "sys/bmfpga.h"
#include "cm-api.h"
#include "cm-api-bc.h"
#include "../ssc-genlib/cm-util-bc.h"
#include "../ssc-genlib/genlib.h"
#ifdef DMALLOC
#include "dmalloc.h"
#endif

/* number of errors before we repeat the warning
*/
#define FC0_ERR_COUNT	4

/* cm bmfpga info
 */
volatile bmfpga_master_t *cm_bmfpga;

uint32 cm_debug;
extern cm_t cm;
extern int chassisd_rxif;


/* fan controller address table
 */
const uint8 fanCtrlTbl[FAN_MAX_CTRL] = {SSC_I2C_ADDR_FAN_CTRL_0,
                                        SSC_I2C_ADDR_FAN_CTRL_1};
/* fan controller register table
 */
const uint8 fanTachMax[FAN_MAX_CTRL] = {FAN_0_MAX_TACH_REG,
                                        FAN_1_MAX_TACH_REG};
/* fan failure count
 */
uint32 fanFailCount = 0;

/*
 * convert a fan alarm status register value to a string.
 */
static char *
fan_alarm_msg(int status)
{
	switch (status) {
		case FAN_ALARM_STATUS_MAX:
			return ("maximum output alarm");
		case FAN_ALARM_STATUS_MIN:
			return ("minimum output alarm");
		case FAN_ALARM_STATUS_TACH:
			return ("tachometer overflow alarm");
		case FAN_ALARM_STATUS_GPIO1:
			return ("GPIO1 alarm");
		case FAN_ALARM_STATUS_GPIO2:
			return ("GPIO2 alarm");
		default:
			return ("unknown");
	}
}

/*++

Routine Description:
    This routine sends a CPU event to the event manager

Arguments:
    slot - source slot
    cpu - source cpu
    state - operational state (EVENT_STATE_UP, EVENT_STATE_DOWN)

Return Value:
    none

--*/
static void
cmSendCpuEvent_bc(uint32 slot, uint32 cpu, int state)
{
    nfx_eventMsg_t eventMsg;
	int rc;

    bzero(&eventMsg, sizeof(eventMsg));
    eventMsg.msgType = NFX_EVENT_POST;
    eventMsg.needAck = FALSE;
    eventMsg.eventScope = NFX_EVENT_LOCAL;
    eventMsg.eventType = NFX_EVENT_CPU;
    eventMsg.param.cpuEvent.slot = slot;
    eventMsg.param.cpuEvent.cpu = cpu;
    eventMsg.param.cpuEvent.state = state;
    rc = sendAgileMsg(chassisd_rxif, &eventMsg, sizeof(nfx_eventMsg_t),
					  "eventd", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE, 0,0,0);
	if (rc < 0) {
        E_LOG(class_1, err_s, eee_getApplicationId("cm"), 0, 0, 0,
              ("CM: %s: Error sending agile msg to eventd %d.\n",
			  __FUNCTION__, rc));
	}
}

/*++

Routine Description:
    This routine sends a power supply event to the event manager

Arguments:
    which - power supply id
    state - operational state (EVENT_STATE_UP, EVENT_STATE_DOWN)

Return Value:
    none

--*/
static void
cmSendPowerEvent(uint32 which, int state)
{
    nfx_eventMsg_t eventMsg;
	int rc;

    bzero(&eventMsg, sizeof(eventMsg));
    eventMsg.msgType = NFX_EVENT_POST;
    eventMsg.needAck = FALSE;
    eventMsg.eventScope = NFX_EVENT_LOCAL;
    eventMsg.eventType = NFX_EVENT_POWERSUPPLY;
    eventMsg.param.powerEvent.state = state;
    if (state == EVENT_STATE_UP)
        sprintf(eventMsg.msg_to_gen_autosupport, 
                "Problem cleared on power supply #%d.\n", which);
    else
        sprintf(eventMsg.msg_to_gen_autosupport,
                "Problem detected on power supply #%d. Check power input.\n",
                which);
    rc = sendAgileMsg(chassisd_rxif, &eventMsg, sizeof(nfx_eventMsg_t),
					  "eventd", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE, 0,0,0);
	if (rc < 0) {
        E_LOG(class_1, err_s, eee_getApplicationId("cm"), 0, 0, 0,
              ("CM: %s: Error sending agile msg to eventd %d.\n",
			  __FUNCTION__, rc));
	}
}

/*++

Routine Description:
    This routine sends a fan event to the event manager

Arguments:
    which - fan id
    state - operational state (EVENT_STATE_UP, EVENT_STATE_DOWN)

Return Value:
    none

--*/
static void
cmSendFanEvent(uint32 which, int state)
{
    nfx_eventMsg_t eventMsg;
    uint8 fan_i2c, data = 0;
	int rc;

    if (which == 1) {
        fan_i2c = SSC_I2C_ADDR_FAN_CTRL_0;
    }
    else {
        fan_i2c = SSC_I2C_ADDR_FAN_CTRL_1;
    }
    if (mm_i2c(I2C_READ, SSC_I2C_BUS_FAN_CTRL, fan_i2c,
               FAN_ALARM_REG, 1, &data) != NFX_OK) {
        E_LOG(class_1, warning_s, eee_getApplicationId("cm"), 0, 0, 0,
              ("CM: Couldn't clear fan alarm\n"));
        data = 0;
    }
    bzero(&eventMsg, sizeof(eventMsg));
    eventMsg.msgType = NFX_EVENT_POST;
    eventMsg.needAck = FALSE;
    eventMsg.eventScope = NFX_EVENT_LOCAL;
    eventMsg.eventType = NFX_EVENT_FAN;
    eventMsg.param.fanEvent.state = state;
    if (state == EVENT_STATE_UP)
        sprintf(eventMsg.msg_to_gen_autosupport, 
                "Problem cleared on fan controller #%d\n", which);
    else
        sprintf(eventMsg.msg_to_gen_autosupport, 
                "Problem detected on fan controller #%d, status=%x (%s)\n",
                which, data, fan_alarm_msg(data));
    rc = sendAgileMsg(chassisd_rxif, &eventMsg, sizeof(nfx_eventMsg_t),
					  "eventd", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE, 0,0,0);
	if (rc < 0) {
        E_LOG(class_1, err_s, eee_getApplicationId("cm"), 0, 0, 0,
              ("CM: %s: Error sending agile msg to eventd %d.\n",
			  __FUNCTION__, rc));
	}
}

static void
cmSendFanSpeedEvent(int32 fan, uint32 tach, uint32 temp)
{
    nfx_eventMsg_t eventMsg;
    int rc;

    bzero(&eventMsg, sizeof(eventMsg));
    eventMsg.msgType = NFX_EVENT_POST;
    eventMsg.needAck = FALSE;
    eventMsg.eventScope = NFX_EVENT_LOCAL;
    eventMsg.eventType = NFX_EVENT_FAN;
    snprintf(eventMsg.msg_to_gen_autosupport, sizeof(eventMsg.msg_to_gen_autosupport),  
            "Fan #%d speed fault, tachometer at %d revolutions per second, intake air temperature at %d degrees Celsius.\n", fan, tach, temp);
    rc = sendAgileMsg(chassisd_rxif, &eventMsg, sizeof(nfx_eventMsg_t),
					  "eventd", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE, 0,0,0);
	if (rc < 0) {
        E_LOG(class_1, err_s, eee_getApplicationId("cm"), 0, 0, 0,
              ("CM: %s: Error sending agile msg to eventd %d.\n",
			  __FUNCTION__, rc));
	}
}

/*++

Routine Description:
    This routine gets the BMFPGA version. 

Arguments:
    verStr - pointer to version string returned

Return Value:
    None

--*/
static void
cmGetVersionBMFPGA(uchar8 *verStr)
{
    uint32 val, ver;

    if (mm_readAddr(&val, 4, CM_BMFPGA_STATUS) == NFX_OK) {
        ver = (val >> 16) & 0xff;
        sprintf(verStr, "BMFPGA : Version %d", ver);  
    }
    else {
        sprintf(verStr, "BMFPGA : unknown");  
    }
    return;
}

/*++

Routine Description:
    This routine checks the power supply status and sends alarms if necessary. 

Arguments:
    none

Return Value:
    NFX_OK or NFX_ERR

--*/
static int
cm_checkPowerStatus(void)
{
    uint32 val;
    static unsigned int PS0_err_count = 0,
                        PS1_err_count = 0;
    if (mm_readAddr(&val, 4, CM_BMFPGA_INTR) == NFX_OK) {
        if (val & CM_BMFPGA_PS0_OK) {
            if (PS0_err_count) {
                cmSendPowerEvent(1, EVENT_STATE_UP);
                E_LOG (class_1, notice_s, eee_getApplicationId("cm"), 0, 0, 0,
                       ("CM: Problem cleared on power supply #1\n"));
                PS0_err_count = 0;
            }
        } else {
            if (!PS0_err_count)
                cmSendPowerEvent(1, EVENT_STATE_DOWN);
            if (!(PS0_err_count++ % PS_ERR_COUNT))
                E_LOG (class_1, warning_s, eee_getApplicationId("cm"), 0, 0, 0,
                       ("CM: Problem detected on power supply #1.  "
                        "Check power input.\n"));
        }
        if (val & CM_BMFPGA_PS1_OK) {
            if (PS1_err_count) {
                cmSendPowerEvent(2, EVENT_STATE_UP);
                E_LOG (class_1, notice_s, eee_getApplicationId("cm"), 0, 0, 0,
                       ("CM: Problem cleared on power supply #2\n"));
                PS1_err_count = 0;
            }
        } else {
            if (!PS1_err_count)
                cmSendPowerEvent(2, EVENT_STATE_DOWN);
            if (!(PS1_err_count++ % PS_ERR_COUNT))
                E_LOG (class_1, warning_s, eee_getApplicationId("cm"), 0, 0, 0,
                       ("CM: Problem detected on power supply #2.  "
                        "Check power input.\n"));
        }
        return NFX_OK;
    }
    return NFX_ERR;
}

/*++

Routine Description:
    This routine checks the fan status and sends alarms if necessary. 

Arguments:
    none

Return Value:
    NFX_OK or NFX_ERR

--*/
static int
cm_checkFanStatus(void)
{
    uint32 val;
    static unsigned int Fan0_err_count = 0,
                        Fan1_err_count = 0;

    if (mm_readAddr(&val, 4, CM_BMFPGA_INTR) == NFX_OK) {
        if (val & CM_BMFPGA_FAN0_OK) {
            if (Fan0_err_count) {
                cmSendFanEvent(1, EVENT_STATE_UP);
                E_LOG (class_1, notice_s, eee_getApplicationId("cm"), 0, 0, 0,
                       ("CM: Problem cleared on fan controller #1\n"));
                Fan0_err_count = 0;
            }
        } else {
            if (!Fan0_err_count)
                cmSendFanEvent(1, EVENT_STATE_DOWN);
            if (!(Fan0_err_count++ % FAN_ERR_COUNT))
                E_LOG (class_1, warning_s, eee_getApplicationId("cm"), 0, 0, 0,
                       ("CM: Problem detected on fan controller #1\n"));
        }
        if (val & CM_BMFPGA_FAN1_OK) {
            if (Fan1_err_count) {
                cmSendFanEvent(2, EVENT_STATE_UP);
                E_LOG (class_1, notice_s, eee_getApplicationId("cm"), 0, 0, 0,
                       ("CM: Problem cleared on fan controller #2\n"));
                Fan1_err_count = 0;
            }
        } else {
            if (!Fan1_err_count)
                cmSendFanEvent(2, EVENT_STATE_DOWN);
            if (!(Fan1_err_count++ % FAN_ERR_COUNT))
                E_LOG (class_1, warning_s, eee_getApplicationId("cm"), 0, 0, 0,
                       ("CM: Problem detected on fan controller #2\n"));
        }
        return NFX_OK;
    }
    return NFX_ERR;
}

/*++

Routine Description:
    This routine sets the specified compact flash LED.

Arguments:
    led - CM_CF_DEV1 or CM_CF_DEV2
    state - CM_LED_OFF, CM_LED_RED, CM_LED_GREEN, CM_LED_AMBER

Return Value:
    none

--*/
void
cm_set_CFLED(uint32 led, uint32 state)
{
    uint32 val;

    if (mm_readAddr(&val, 4, CM_MARVELL_GPIO) == NFX_OK) {
        if (led == CM_CF_DEV1) {
            val &= 0xcfffffff;
            val |= ((state & 0x3) << 28);
        }
        else 
        if (led == CM_CF_DEV2) {
            val &= 0x3fffffff;
            val |= ((state & 0x3) << 30);
        }
        mm_writeAddr(&val, 4, CM_MARVELL_GPIO);
    }
}

/*++

Routine Description:
    This routine sets the front panel status OK LED. 

Arguments:
    state - CM_LED_OFF, CM_LED_RED, CM_LED_GREEN, CM_LED_AMBER 

Return Value:
    none

--*/
static void
cm_set_OKLED(uint32 state)
{
    uint32 val;

    if (mm_readAddr(&val, 4, CM_MARVELL_GPIO) == NFX_OK) {
        val &= 0xff9fffff;
        val |= ((state & 0x3) << 21);
        mm_writeAddr(&val, 4, CM_MARVELL_GPIO);
    }
}

/*++

Routine Description:
    This routine sets the front panel status fault LED. 

Arguments:
    state - CM_LED_OFF, CM_LED_RED, CM_LED_GREEN, CM_LED_AMBER 

Return Value:
    none

--*/
static void
cm_set_FLTLED(uint32 state)
{
    uint32 val;

    if (mm_readAddr(&val, 4, CM_MARVELL_GPIO) == NFX_OK) {
        val &= 0xfe7fffff;
        val |= ((state & 0x3) << 23);
        mm_writeAddr(&val, 4, CM_MARVELL_GPIO);
    }
}

/*++

Routine Description:
    This routine checks the CPU status of all modules and sends UP/DOWN
    events to the event manager. For Bobcat, the FAULT LED is set red if
    any cpu is not in runtime state; otherwise, the OK LED is set green.

Arguments:
    none

Return Value:
    none

--*/
static void
cm_cpuPoll(void)
{
    uint32 val, state, cpu_id, fault = 0, wait_crash_dump = 0;
    cm_cpu_t *cpu;
    static unsigned int FC0_err_count = 0;

    if (mm_readAddr(&val, 4, CM_BMFPGA_FP_STAT1) == NFX_OK) {
        for (cpu_id = 0; cpu_id < cm.slot[NFX_SLOTTYPE_NFP].num_cpus; cpu_id++){
            cpu = &cm.slot[NFX_SLOTTYPE_NFP].cpu[cpu_id];
            state = val & BMFPGA_SLOT_STATUS_SLICE_STATE_MASK; 
            if ((cpu->state != CM_SLICE_STATE_RUNTIME) &&
                (state == CM_SLICE_STATE_RUNTIME)) {
                if (cpu->state != CM_SLICE_STATE_RUNTIME_REBOOT) {
                    CM_DBG_PRINTF(("FP Module, CPU %d UP\n", cpu_id));
                    cmSendCpuEvent_bc(NFX_SLOTTYPE_NFP, cpu_id, EVENT_STATE_UP);
                    cpu->downtime = 0;
                }
            }
            else { 
                if ((cpu->state == CM_SLICE_STATE_RUNTIME)
                    && (state != CM_SLICE_STATE_RUNTIME)) {
                    CM_DBG_PRINTF(("FP Module, CPU %d DOWN\n", cpu_id));
                    cmSendCpuEvent_bc(NFX_SLOTTYPE_NFP, cpu_id, EVENT_STATE_DOWN);
                }
            }
            cpu->state = state;
            if (cpu->state != CM_SLICE_STATE_RUNTIME) {
                fault = 1;
                cpu->downtime += CM_SHORT_POLL_TIME_SECS;
                if ((cpu->state == CM_SLICE_STATE_RESET_WAIT) || 
                    (cpu->state == CM_SLICE_STATE_CRASH_DONE) ||
                    (cpu->downtime >= CM_DOWNTIME_MAX_MINUTES * SEC_PER_MIN)) {
                    if (cmGetAutoreboot_flag()) {
                        cmResetMacs();
                        cm_delay_sec_bc(2);
                        utils_reboot_system("CM: FP module is down,"
							" rebooting system.", "cm", 0);
                    }
                }
                else 
                if (cpu->state == CM_SLICE_STATE_RUNTIME_CRASHED) {
                    wait_crash_dump = 1;
                }
            }
        }
    }
    else {
        CM_DBG_PRINTF(("Couldn't check CPU status for FP module."));
    }

    if (mm_readAddr(&val, 4, CM_BMFPGA_FC0_STAT1) == NFX_OK) {
        cpu_id = 0;
        cpu = &cm.slot[NFX_SLOTTYPE_FC].cpu[cpu_id];
        state = val & BMFPGA_SLOT_STATUS_SLICE_STATE_MASK; 

        if ((cpu->state != CM_SLICE_STATE_RUNTIME) &&
            (state == CM_SLICE_STATE_RUNTIME)) {
            if (cpu->state != CM_SLICE_STATE_RUNTIME_REBOOT) {
                CM_DBG_PRINTF(("SP Module, CPU %d UP\n", cpu_id));
                cmSendCpuEvent_bc(NFX_SLOTTYPE_FC, cpu_id, EVENT_STATE_UP);
                FC0_err_count = 0; // clear counter
            }
        } else { 
            if ((cpu->state == CM_SLICE_STATE_RUNTIME)
                && (state != CM_SLICE_STATE_RUNTIME)) {
                CM_DBG_PRINTF(("SP Module, CPU %d State was [%d] now [%d]\n", 
                               cpu_id, cpu->state, state ));
                E_LOG(class_1, warning_s, eee_getApplicationId("cm"),0,0,0,
                      ("SP Module, CPU %d State was [%d] now [%d], possible reboot.\n", 
                       cpu_id, cpu->state, state ));
            }
        }
        cpu->state = state; /* save the last known state */

        if ((cpu->state != CM_SLICE_STATE_RUNTIME) &&
            (FC0_err_count++ >= FC0_ERR_COUNT )) {
            /* Only send the event once */
            if(FC0_err_count == FC0_ERR_COUNT ) {
                cmSendCpuEvent_bc(NFX_SLOTTYPE_FC, cpu_id, EVENT_STATE_DOWN);
            }
            if (!wait_crash_dump) {
                if (cmGetAutoreboot_flag()) {
                    cmResetMacs();
                    cm_delay_sec_bc(2);
                    utils_reboot_system("Node going down for reboot!"
						"(chassisd: because FC module is down).", "cm", 0);
                }
            }
            fault = 1;
        }
    }
    else {
        CM_DBG_PRINTF(("Couldn't check CPU status for SP module."));
    }

    if (fault == 0) {
        cm_set_OKLED(CM_LED_GREEN);
        cm_set_FLTLED(CM_LED_OFF);
    }
    else {
        cm_set_FLTLED(CM_LED_RED);
        cm_set_OKLED(CM_LED_OFF);
    }
}

/*++

Routine Description:
    This is the periodic polling routine which is called at fixed time
    intervals to check on system status and etc.  Message polling is done 
    elsewhere.

Arguments:
    none

Return Value:
    none

--*/
static void
cm_poll(void)
{

    /* 
     * check power supply status
     * check fan status
     * check temperature alarm threshold
     */

    if (cm_checkPowerStatus() != NFX_OK) {
        E_LOG(class_1, warning_s, eee_getApplicationId("cm"), 0, 0, 0,
             ("CM: Couldn't check power supply status\n"));
    }
    if (cm_checkFanStatus() != NFX_OK) {
        E_LOG(class_1, warning_s, eee_getApplicationId("cm"), 0, 0, 0,
             ("CM: Couldn't check fan status\n"));
    }
    if (cm_checkTempStatus() != NFX_OK) {
        E_LOG(class_1, warning_s, eee_getApplicationId("cm"), 0, 0, 0,
             ("CM: Couldn't check system temperature.\n"));
    }
}

/*++

Routine Description:
    This routine clears the fan alarm registers.

Arguments
    none

Return Value
    NFX_OK or NFX_ERR

--*/
static int
cmFanClrAlarm(void) 
{
    uint8 data;

    if (mm_i2c(I2C_READ, SSC_I2C_BUS_FAN_CTRL, SSC_I2C_ADDR_FAN_CTRL_0,
               FAN_ALARM_REG, 1, &data)
        != NFX_OK) {
        return NFX_ERR;
    }
    if (mm_i2c(I2C_READ, SSC_I2C_BUS_FAN_CTRL, SSC_I2C_ADDR_FAN_CTRL_1,
               FAN_ALARM_REG, 1, &data)
        != NFX_OK) {
        return NFX_ERR;
    }
    return NFX_OK;
}

/*++

Routine Description:
    This routine is called when a fan fault is detected to attempt to run
    the fans at full speed and clear the fan alarm registers.

Arguments:
    none

Return Value:
    NFX_OK or NFX_ERR
 
--*/
static int
cmFanFailed(void)
{
    uint32 data;

    if (++fanFailCount >= FAN_MAX_FAIL_COUNT) {
        if (mm_readAddr(&data, 4, CM_BMFPGA_STATUS) == NFX_OK) {
            data |= CM_BMFPGA_FAN_FULL;
            if (mm_writeAddr(&data, 4, CM_BMFPGA_INTR) != NFX_OK) {
                E_LOG(class_1, warning_s, eee_getApplicationId("cm"), 0, 0, 0,
                     ("CM: Couldn't update fan controller.\n"));
                return NFX_ERR;
            } 
            if (cmFanClrAlarm() != NFX_OK) {
                return NFX_ERR;
            }
        } 
    }
    return NFX_OK;
}

/*++

Routine Description:
    This routine monitors the fan controllers and follows this procedure:

    Go thru TACH register for each fan controller.
    If any one of the TACH value is out of range,
    or temperature is too high, assume a fan is bad.
    After 5 times, turn all fans to FULL power.
 
    If TACH values are in range & temperture returns
    to normal, disable FULL power and enable close loop
    control.

Arguments:
    none

Return Value:
    none
 
--*/
static void
cmFanMonitor(void)
{
    int32   i, j;
    int32   ctrl_addr;
    int32   tach_max;
    int32   tach_reg;
    int32   fan = 0;
    uint32  temp_val=0, data, min_rev, max_rev;
    uchar8  tach_val;
    uint8   byte;
    static unsigned int fan_err_count=0;

    if (mm_readAddr(&data, 4, CM_BMFPGA_INTR) != NFX_OK) {
        E_LOG(class_1, warning_s, eee_getApplicationId("cm"), 0, 0, 0,
             ("CM: Couldn't monitor fan speed status\n"));
        return;
    }

    min_rev = (data & CM_BMFPGA_FAN_FULL) ? FAN_HIGH_MIN_REV_COUNT
                                          : FAN_LOW_MIN_REV_COUNT;
    max_rev = (data & CM_BMFPGA_FAN_FULL) ? FAN_HIGH_MAX_REV_COUNT
                                          : FAN_LOW_MAX_REV_COUNT;

    /* Read tach register values from FAN Controllers 0 & 1
     */
    for (i=0; i < FAN_MAX_CTRL; i++)
    {
        ctrl_addr = fanCtrlTbl[i];  // set ctrl I2C address
        tach_max  = fanTachMax[i];  // set # of tach reg to read
        tach_reg  = FAN_TACH0_REG;  // set base tach reg addr

        for (j=0; j < tach_max; j++)
        {
            if (mm_i2c(I2C_READ, SSC_I2C_BUS_FAN_CTRL, ctrl_addr,
                       tach_reg, 1, &tach_val) != NFX_OK) {
                E_LOG(class_1, warning_s, eee_getApplicationId("cm"), 0, 0, 0,
                     ("CM: Couldn't monitor fan status\n"));
                return;
            }
#if 0
            CM_DBG_PRINTF(("fan[%d] ctrl=0x%0x tach=0x%x val=0x%0x\n",
                          fan, ctrl_addr, tach_reg, tach_val));
#endif
            /* check if fan speed is within range
             */
            if ((tach_val < min_rev) || (tach_val > max_rev)) {
                cm_read_temperature_bc(&temp_val);
                if (!fan_err_count)
                    cmSendFanSpeedEvent(fan+1, tach_val, INTAKE_TEMP(temp_val));
                if (!(fan_err_count++ % FAN_ERR_COUNT))
                    E_LOG(class_1, warning_s, eee_getApplicationId("cm"), 0,0,0,
                          ("CM: Fan #%d speed fault, tachometer at %d rps, "
                           "intake air temperature at %d degrees Celsius.\n", 
                           fan+1, tach_val, INTAKE_TEMP(temp_val)));
                cmFanFailed();
                return;
            }
            tach_reg += FAN_REG_OFFSET;
            fan++;
        }
    }

    if (fan_err_count) {
        E_LOG(class_1, notice_s, eee_getApplicationId("cm"), 0,0,0,
              ("CM: Fan speed returned to normal range.\n"));
        fan_err_count = 0;
    }
    if (cm_read_temperature_bc(&temp_val) != NFX_OK) {
        E_LOG(class_1, warning_s, eee_getApplicationId("cm"), 0, 0, 0,
             ("CM: Couldn't check system temperature.\n"));
        cmFanFailed();
        return;
    }
    /* temperature too high, treat as fan failure.
     */
    if (temp_val >= FAN_MAX_CHASSIS_TEMP)
    {
        cmFanFailed();
        return;
    }

    /* temperature & TACH values are in range.
     * Disable FULL on and enable close loop control.
     */
    if (mm_readAddr(&data, 4, CM_BMFPGA_INTR) != NFX_OK) {
        return;
    }

    if ((data & CM_BMFPGA_FAN_FULL) && (temp_val <= (FAN_MAX_CHASSIS_TEMP-3)))
    {
        cmFanClrAlarm();

        if (mm_readAddr(&data, 4, CM_BMFPGA_INTR) == NFX_OK) {
            data &= ~CM_BMFPGA_FAN_FULL;
            if (mm_writeAddr(&data, 4, CM_BMFPGA_INTR) != NFX_OK) {
                E_LOG(class_1, warning_s, eee_getApplicationId("cm"), 0, 0, 0,
                     ("CM: Couldn't update fan controller.\n"));
                return;
            }
        }

        cm_delay_us_bc(100);

        /* Need to clear alarm before enabling closed loop.
         */
        cmFanClrAlarm();

        byte = FAN_CONFIG_CLOSE_LOOP | FAN_CONFIG_12VOLT | FAN_CONFIG_DIV_BY_4;
        if (mm_i2c(I2C_WRITE, SSC_I2C_BUS_FAN_CTRL, SSC_I2C_ADDR_FAN_CTRL_0,
                   FAN_CONFIG_REG, 1, &byte) != NFX_OK) {
            E_LOG(class_1, warning_s, eee_getApplicationId("cm"), 0, 0, 0,
                 ("CM: Couldn't update fan controller #1.\n"));
        }
        byte = FAN_CONFIG_CLOSE_LOOP | FAN_CONFIG_12VOLT | FAN_CONFIG_DIV_BY_4;
        if (mm_i2c(I2C_WRITE, SSC_I2C_BUS_FAN_CTRL, SSC_I2C_ADDR_FAN_CTRL_1,
                   FAN_CONFIG_REG, 1, &byte) != NFX_OK) {
            E_LOG(class_1, warning_s, eee_getApplicationId("cm"), 0, 0, 0,
                 ("CM: Couldn't update fan controller #2.\n"));
        }
    }
    return;
}

/*++

Routine Description:
    This is the initialization routine for the chassis manager called once
    during startup.

Arguments:
    none

Return Value:
    none

--*/
static void
chassisd_init(void)
{
    uint32   module, cpu, cf;
    uchar8 ver[CM_VER_STRING_LEN];
    seepPCB_t seep;

    for (module = 0; module <= 2; module++) {
        cm.slot[module].present = 1;
        cm.slot[module].temp_valid = 0;
        cm.slot[module].power_valid = 0;
        cm.slot[module].control = 0xFF;
        cm.slot[module].downtime = 0;
        for (cpu = 0; cpu < CM_NUM_CPUS; cpu++) {
            cm.slot[module].cpu[cpu].downtime = 0;
            cm.slot[module].cpu[cpu].outtime = 0;
        }
    }
    cm.slot[0].type = CM_CARD_TYPE_SSC;
    cm.slot[0].num_cpus = 1;
    cm.slot[0].cpu[0].state = CM_SLICE_STATE_RUNTIME;
    cm.slot[1].type = CM_CARD_TYPE_BOBCAT;
    cm.slot[1].num_cpus = 4;
    cm.slot[2].type = CM_CARD_TYPE_FCNIM;
    cm.slot[2].num_cpus = 1;
    cmGetVersionSSC(&ver[0]);
    strcpy(cm.slot[NFX_SLOTTYPE_SSC].cpu[0].prom_ver, ver); 
    cmGetVersionBMFPGA(&ver[0]);
    strcpy(cm.slot[NFX_SLOTTYPE_SSC].bmfpga_ver, ver);
    cmGetBootDevInfo_bc(&ver[0], &cf);
    strcpy(cm.slot[NFX_SLOTTYPE_SSC].boot_dev, ver);
    if (cm_read_seep_bc((void *)&seep, 0, sizeof(seepPCB_t)) == NFX_OK) {
        sprintf(cm.slot[NFX_SLOTTYPE_SSC].chassis_macAddr, 
               "MAC addr: %c%c:%c%c:%c%c:%c%c:%c%c:%c%c", 
               seep.mac_addr[0], seep.mac_addr[1],
               seep.mac_addr[2], seep.mac_addr[3],
               seep.mac_addr[4], seep.mac_addr[5],
               seep.mac_addr[6], seep.mac_addr[7],
               seep.mac_addr[8], seep.mac_addr[9],
               seep.mac_addr[10], seep.mac_addr[11]);
        sprintf(cm.slot[NFX_SLOTTYPE_SSC].chassis_serialNo, 
               "System serial number: %s", seep.serial_num);
        sprintf(cm.slot[NFX_SLOTTYPE_SSC].hw_ver,
               "Board revision: %s", seep.board_revision);
        sprintf(cm.slot[NFX_SLOTTYPE_SSC].hw_serialNo, 
               "Board serial number: %s", seep.board_serial_num);
    }
    else {
        strcpy(cm.slot[NFX_SLOTTYPE_SSC].chassis_macAddr, 
              "MAC addr: unknown");
        strcpy(cm.slot[NFX_SLOTTYPE_SSC].chassis_serialNo, 
              "System serial number: unknown");
        strcpy(cm.slot[NFX_SLOTTYPE_SSC].hw_ver, 
              "Board revision: unknown");
        strcpy(cm.slot[NFX_SLOTTYPE_SSC].hw_serialNo,
              "Board serial number: unknown");
    }
    cm_update_CFLED();
}

/*++

Routine Description:
    This routine initializes the mmap interface and i2c semaphore. Then it
    loops forever monitoring the power supplies, fans, temperature and FP
    processor runtime status.

Arguments:
    none

Return Value:
    none
 
--*/
void chassisd_run(void)
{
    struct timeval tv, pv_long, pv_short;
    int rc;

    rc = cmSetWatchDog(CM_WATCHDOG_TIMEOUT);
    if (rc != 0) {
        E_LOG(class_1, err_s, eee_getApplicationId("cm"), 0, 0, 0,
             ("CM: Couldn't set watchdog timer.\n"));
    }
    if (mm_init() != 0) {
        exit(1);
    }
    if (cm_init_i2c_access() != NFX_OK) {
        E_LOG(class_1, err_s, eee_getApplicationId("cm"), 0, 0, 0,
             ("CM: Couldn't initialize i2c semaphore.\n"));
    }
    chassisd_init();
    pv_long.tv_usec = 0;
    pv_long.tv_sec = 0;
    pv_short.tv_usec = 0;
    pv_short.tv_sec = 0;
    while (1) {
        chassisd_rcvMsg();
        cmSetWatchDog(CM_WATCHDOG_TIMEOUT);
        /* if certain time interval has elapsed then poll for status
         */
        if (gettimeofday(&tv, NULL) == 0) {
            if (pv_long.tv_sec + CM_LONG_POLL_TIME_SECS < tv.tv_sec) {
                cm_poll();
                cmFanMonitor();
                pv_long = tv;
            }
            if (pv_short.tv_sec + CM_SHORT_POLL_TIME_SECS < tv.tv_sec) {
                cm_cpuPoll();
                pv_short = tv;
            }
            /* resync to handle the case if time is set backwards
             */
            if (pv_short.tv_sec > tv.tv_sec) {
                pv_short = tv;
                pv_long = tv;
            }
        }
    }
}
#endif  /* SSC */
