/*
 *-----------------------------------------------------------------
 *
 * Name: chassisd-cg.c
 *
 * Copyright (C) 2007, Onstor, Inc.
 *
 * Description:
 *
 *   Chassis manager daemon code for Cougar.
 *
 * Created by:  Rendell Fong
 *
 * Date Created:    11/01/07
 *
 *-----------------------------------------------------------------
 */
#ifdef SSC

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>
#include "../sm-anpssc/anpssc-api.h"
#include "nfx-incl.h"
#include "../sm-elog/elog-api.h"
#include "../sm-utils/sys-utils-api.h"
#include "../sm-event/event-api.h"
#include "../sm-seep/seep.h"
#include "cm.h"
#include "sys/bmfpga.h"
#include "cm-api.h"
#include "cm-api-cg.h"
#include "../ssc-genlib/cm-util-bc.h"
#include "../ssc-genlib/genlib.h"
#ifdef DMALLOC
#include "dmalloc.h"
#endif


/* cm bmfpga info
 */
volatile bmfpga_master_t *cm_bmfpga;

uint32 cm_debug;
extern cm_t cm;
extern int chassisd_rxif;

/* Chassis SEEP
 */
seepChassis_t chsSeep;

/* power supply address/status tables
 */
const uint8 psRegTbl[PS_MODULE_CNT] =
    {SSC_I2C_ADDR_PS0, SSC_I2C_ADDR_PS1};
const uint32 psOkBit[PS_MODULE_CNT] =
    {CM_BMFPGA_PS0_OK, CM_BMFPGA_PS1_OK};

/* fan controller address table
 */
const uint8 fanCtrlTbl[FAN_MODULE_CNT][FAN_MAX_CTRL] = {
    {SSC_I2C_ADDR_FAN_CTRL_0, SSC_I2C_ADDR_FAN_CTRL_1},
    {SSC_I2C_ADDR_FAN_CTRL_0, SSC_I2C_ADDR_FAN_CTRL_1},
    {SSC_I2C_ADDR_FAN_CTRL_0, SSC_I2C_ADDR_FAN_CTRL_1}
};

/* CPU name table
 */
const char *cpuList[NFX_NFP_FP_LAST+1] = {
    CM_CPU_TYPE_TXRX0_STR,
    CM_CPU_TYPE_TXRX1_STR,
    CM_CPU_TYPE_FP0_STR,
    CM_CPU_TYPE_FP1_STR,
    CM_CPU_TYPE_FP2_STR,
    CM_CPU_TYPE_FP3_STR };

/*++

Routine Description:
    This routine translates a fan alarm status register value to a
    descriptive string.

Arguments:
    status - alarm status register value

Return Value:
    address of alarm description string

--*/
static char *
fan_alarm_msg(int status)
{
    static char desc[80];
    uint32 temp = 0;
    int rc;

    if ((status & FAN_ALARM_STATUS_MAX)) {
        rc = cm_read_temperature_bc(&temp);
        if (rc == NFX_OK) {
            snprintf(desc, sizeof(desc), "maximum output - %u degrees Celsius",
                     INTAKE_TEMP(temp));
            return desc;
        } else
            return "maximum output - N/A degrees Celsius";
    } else if ((status & FAN_ALARM_STATUS_MIN)) {
        rc = cm_read_temperature_bc(&temp);
        if (rc == NFX_OK) {
            snprintf(desc, sizeof(desc), "minimum output - %u degrees Celsius",
                     INTAKE_TEMP(temp));
            return desc;
        } else
            return "minimum output - N/A degrees Celsius";
    } else if ((status & FAN_ALARM_STATUS_TACH)) {
        return "tach overflow";
    } else if ((status & FAN_ALARM_ENABLE_GPIO1)) {
        return "GPIO1";
    } else {
        return "unknown";
    }
}

/*++

Routine Description:
    This routine translates power supply alert event register values to a
    descriptive string.

Arguments:
    evtData - address of alert status register values

Return Value:
    address of alert description string

--*/
static char *
ps_alert_msg(uint8 *evtData)
{

    /* Check for all critical event types
     */
    if (evtData[0] & PS_CRIT_EVENT) {
        if (evtData[1] & PS_CRIT1_FAIL)
            return "critical: hard failure";
        else if (evtData[1] & PS_CRIT1_12V_OVRPWR)
            return "critical: +12V out over power";
        else if (evtData[1] & PS_CRIT1_OVRTEMP)
            return "critical: over temperature";
        else if (evtData[1] & PS_CRIT1_AC_LOST)
            return "critical: AC lost";
        else if (evtData[2] & PS_CRIT2_12V_OVRCUR)
            return "critical: over current";
        else
            return "critical: unknown";
    } 

    /* Check for all warning event types
     */
    if (evtData[0] & PS_WARN_EVENT) {
        if (evtData[3] & PS_WARN1_TEMP)
            return "warning: temperature";
        else if (evtData[3] & PS_WARN1_SLOW_FAN)
            return "warning: slow fan";
        else if (evtData[3] & PS_WARN1_AC_INCUR)
            return "warning: AC input current";
        else if (evtData[3] & PS_WARN1_12V_PWR)
            return "warning: +12V out power";
        else if (evtData[4] & PS_WARN2_12V_OVRCUR)
            return "warning: over current";
        else
            return "warning: unknown";
    }

    /* Unknown event type
     */
    return "type?: unknown";
}

/*++

Routine Description:
    This routine sends a CPU event to the event manager

Arguments:
    slot - source slot
    cpu - source cpu
    state - operational state (EVENT_STATE_UP, EVENT_STATE_DOWN)

Return Value:
    none

--*/
static void
cmSendCpuEvent_cg(uint32 slot, uint32 cpu, int state)
{
    nfx_eventMsg_t eventMsg;
	int rc;

    bzero(&eventMsg, sizeof(eventMsg));
    eventMsg.msgType = NFX_EVENT_POST;
    eventMsg.needAck = FALSE;
    eventMsg.eventScope = NFX_EVENT_LOCAL;
    eventMsg.eventType = NFX_EVENT_CPU;
    eventMsg.param.cpuEvent.slot = slot;
    eventMsg.param.cpuEvent.cpu = cpu;
    eventMsg.param.cpuEvent.state = state;

    rc = sendAgileMsg(chassisd_rxif, &eventMsg, sizeof(nfx_eventMsg_t),
					  "eventd", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE, 0,0,0);
	if (rc < 0) {
        E_LOG(class_1, err_s, eee_getApplicationId("cm"), 0, 0, 0,
              ("CM: %s: Error sending agile msg to eventd %d.\n",
			  __FUNCTION__, rc));
	}
}

/*++

Routine Description:
    This routine sends a power supply event to the event manager

Arguments:
    module - power supply id
    state - operational state (EVENT_STATE_UP, EVENT_STATE_DOWN)
    evtData - power supply event data

Return Value:
    none

--*/
static void
cmSendPowerEvent(uint32 module, int state, uint8 *evtData)
{
    nfx_eventMsg_t eventMsg;
	int rc;

    bzero(&eventMsg, sizeof(eventMsg));
    eventMsg.msgType = NFX_EVENT_POST;
    eventMsg.needAck = FALSE;
    eventMsg.eventScope = NFX_EVENT_LOCAL;
    eventMsg.eventType = NFX_EVENT_POWERSUPPLY;
    eventMsg.param.powerEvent.state = state;
    if (state == EVENT_STATE_UP) {
        snprintf(eventMsg.msg_to_gen_autosupport, EVENT_MAX_MAIL_SUB-1,
                 "Problem cleared on power supply #%d.\n", module);
    } else if (evtData) {
        snprintf(eventMsg.msg_to_gen_autosupport, EVENT_MAX_MAIL_SUB-1,
                 "Problem detected on power supply #%d, "
                 "events=[%02x:%02x:%02x:%02x:%02x] (%s)\n",
                 module, evtData[0], evtData[1], evtData[2], evtData[3],
                 evtData[4], ps_alert_msg(evtData));
    } else {
        snprintf(eventMsg.msg_to_gen_autosupport, EVENT_MAX_MAIL_SUB-1,
                 "Problem detected on power supply #%d.\n", module);
    }
    strcpy(eventMsg.param.powerEvent.power_msg, 
           eventMsg.msg_to_gen_autosupport);

    rc = sendAgileMsg(chassisd_rxif, &eventMsg, sizeof(nfx_eventMsg_t),
					  "eventd", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE, 0,0,0);
	if (rc < 0) {
        E_LOG(class_1, err_s, eee_getApplicationId("cm"), 0, 0, 0,
              ("CM: %s: Error sending agile msg to eventd %d.\n",
			  __FUNCTION__, rc));
	}
}

/*++

Routine Description:
    This routine sends a fan event to the event manager

Arguments:
    module - fan module (0-2)
    fan    - fan number (0-1) within module
    state  - operational state (EVENT_STATE_UP, EVENT_STATE_DOWN)
    alarm_val - alarm register value
    tach_val  - tach0 register value

Return Value:
    none

--*/
static void
cmSendFanEvent(uint32 module, uint32 fan, int state, uint8 alarm_val,
               uint8 tach_val)
{
    nfx_eventMsg_t eventMsg;
	int rc;

    bzero(&eventMsg, sizeof(eventMsg));
    eventMsg.msgType = NFX_EVENT_POST;
    eventMsg.needAck = FALSE;
    eventMsg.eventScope = NFX_EVENT_LOCAL;
    eventMsg.eventType = NFX_EVENT_FAN;
    eventMsg.param.fanEvent.state = state;
    if (state == EVENT_STATE_UP) {
        snprintf(eventMsg.msg_to_gen_autosupport, EVENT_MAX_MAIL_SUB-1,
                 "Problem cleared on fan module #%u fan #%u.\n", module, fan);
    } else {
        snprintf(eventMsg.msg_to_gen_autosupport, EVENT_MAX_MAIL_SUB-1,
                 "Problem detected on fan module #%u fan #%u, "
                 "tach=%u rps, alarm=%02x (%s)\n",
                 module, fan, tach_val, alarm_val, fan_alarm_msg(alarm_val));
    }
    strcpy(eventMsg.param.fanEvent.fan_msg, 
           eventMsg.msg_to_gen_autosupport);

    rc = sendAgileMsg(chassisd_rxif, &eventMsg, sizeof(nfx_eventMsg_t),
					  "eventd", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE, 0,0,0);
	if (rc < 0) {
        E_LOG(class_1, err_s, eee_getApplicationId("cm"), 0, 0, 0,
              ("CM: %s: Error sending agile msg to eventd %d.\n",
			  __FUNCTION__, rc));
	}
}

/*++

Routine Description:
    This routine gets the BMFPGA version. 

Arguments:
    verStr - pointer to version string returned

Return Value:
    None

--*/
static void
cmGetVersionBMFPGA(uchar8 *verStr)
{
    uint32 val, ver;

    if (mm_readAddr(&val, sizeof(val), CM_BMFPGA_STATUS) == NFX_OK) {
        ver = (val >> 24) & 0xff;
        sprintf(verStr, "BMFPGA : Version %d", ver);  
    }
    else {
        sprintf(verStr, "BMFPGA : unknown");  
    }
    return;
}

/*++

Routine Description:
    This routine checks the power supply status and sends event notification
    messages if necessary.  Only the most severe event is reported if multiple 
    errors are indicated.

Arguments:
    bdStatus - board status

Return Value:
    NFX_OK

--*/
static int
cm_checkPowerStatus(uint32 bdStatus)
{
    static uint32 PS_err_count[PS_MODULE_CNT] = {0,0};
    uint32 module;
    uint8 evtData[5];
    int rc;

    for (module = 0; module < PS_MODULE_CNT; module++) {
        /* When pwr supply module is ok, check if it's recovered from an error
         */
        if ((bdStatus & psOkBit[module])) {
            if (PS_err_count[module]) {
                cmSendPowerEvent(module, EVENT_STATE_UP, NULL);
                PS_err_count[module] = 0;

                /* Clear PS module event registers
                 */
                if (cm_lock_mezz_i2c_bus() != NFX_OK) {
                    E_LOG(class_1, notice_s, eee_getApplicationId("cm"),0,0,0,
                       ("Can't lock mezzanine I2C and access pwr #%u status.\n",
                        module));
                    return NFX_ERR;
                }
                rc = mm_i2c_mezz(I2C_WRITE, SSC_I2C_BUS_PS,
                                 psRegTbl[module], PS_EVENT_STATUS_REG+1,
                                 sizeof(evtData)-1, evtData);
                if (rc != NFX_OK) {
                    E_LOG(class_1, notice_s, eee_getApplicationId("cm"),0,0,0,
                          ("Can't clear pwr #%u event status.\n", module));
                }
                cm_unlock_mezz_i2c_bus();
            }
        } else {
            /* Interrogate PS module to identify the failure
             */
            if (cm_lock_mezz_i2c_bus() != NFX_OK) {
                E_LOG(class_1, notice_s, eee_getApplicationId("cm"),0,0,0,
                   ("Can't lock mezzanine I2C and access pwr #%u status.\n",
                    module));
                return NFX_ERR;
            }
            rc = mm_i2c_mezz(I2C_READ, SSC_I2C_BUS_PS,
                             psRegTbl[module], PS_EVENT_STATUS_REG,
                             sizeof(evtData), evtData);
            if (rc != NFX_OK) {
                E_LOG(class_1, notice_s, eee_getApplicationId("cm"),0,0,0,
                      ("Can't read power supply #%u event status.\n",
                       module));
            }
            cm_unlock_mezz_i2c_bus();

            if (!PS_err_count[module]) {
                cmSendPowerEvent(module, EVENT_STATE_DOWN,
                                 (rc == NFX_OK) ? evtData : NULL);
            } else if (!(PS_err_count[module] % PS_ERR_COUNT)) {
                if (rc == NFX_OK)
                    E_LOG(class_1, warning_s, eee_getApplicationId("cm"),0,0,0,
                      ("CM: Problem detected on power supply #%u, "
                       "events=[%02x:%02x:%02x:%02x:%02x] (%s)\n", module,
                       evtData[0], evtData[1], evtData[2], evtData[3],
                       evtData[4], ps_alert_msg(evtData)));
                else
                    E_LOG(class_1, warning_s, eee_getApplicationId("cm"),0,0,0,
                      ("CM: Problem detected on power supply #%u\n", module));
            }
            PS_err_count[module]++;
        }
    }

    return NFX_OK;
}

/*++

Routine Description:
    This routine checks the fan status and sends alarms if necessary. 

Arguments:
    bdStatus - board status

Return Value:
    NFX_OK

--*/
static int
cm_checkFanStatus(uint32 bdStatus)
{
    static boolean fanSummaryOk = TRUE;
    static uint32 fan_err_count[FAN_MODULE_CNT][FAN_MAX_CTRL] =
                     {{0,0},{0,0},{0,0}};
    static uint32 mod_err_count[FAN_MODULE_CNT] = {0,0,0};
    boolean module_err;
    uint32 module, fan, val, stat_val;
    uint8  byte, alarm_val, tach_val;
    uint8  min_rev, max_rev;

    /* When fan summary status is still okay since last check, exit 
     */
    if (((bdStatus & CM_BMFPGA_FANS_OK) != 0) && (fanSummaryOk)) {
        return NFX_OK;
    }

    /* Lock mezzanine I2C bus to prevent concurrent access via other blade.
     */
    if (cm_lock_mezz_i2c_bus() != NFX_OK) {
        E_LOG(class_1, notice_s, eee_getApplicationId("cm"),0,0,0,
              ("Can't lock mezzanine I2C and access fan status.\n"));
        return NFX_ERR;
    }

    /* Update fan summary status for next check
     */
    fanSummaryOk = (bdStatus & CM_BMFPGA_FANS_OK) ? TRUE : FALSE;

    min_rev = FAN_LOW_MIN_REV_COUNT;
    max_rev = FAN_LOW_MAX_REV_COUNT;

    /* Check status of each fan module
     */
    for (module = 0; module < FAN_MODULE_CNT; module++) {
        /* Set I2C mux for communication with this fan module
         */
        byte = FAN_MUX_CHAN_ENABLE | module;
        if (mm_i2c_mezz(I2C_WRITE, SSC_I2C_BUS_FAN_CTRL,
                        SSC_I2C_FAN_MUX_CTRL, FAN_MUX_CTRL_REG,
                        1, &byte) != NFX_OK) {
            E_LOG(class_1, warning_s, eee_getApplicationId("cm"),
                  0, 0, 0, ("CM: Couldn't setup fan mux\n"));
            continue;
        }

        module_err = FALSE;

        /* check status for each fan in the module
         */
        for (fan=0; fan < FAN_MAX_CTRL; fan++) {
            /* Read fan controller alarm status
             */
            if (mm_i2c_mezz(I2C_READ, SSC_I2C_BUS_FAN_CTRL,
                            fanCtrlTbl[module][fan], FAN_ALARM_REG,
                            1, &alarm_val) != NFX_OK) {
                E_LOG(class_1, warning_s, eee_getApplicationId("cm"), 0, 0, 0,
                      ("CM: Couldn't get fan #%u alert status\n", fan));
                continue;
            }

            /* Read fan controller tach speed
             */
            if (mm_i2c_mezz(I2C_READ, SSC_I2C_BUS_FAN_CTRL,
                            fanCtrlTbl[module][fan], FAN_TACH0_REG,
                            1, &tach_val) != NFX_OK) {
                E_LOG(class_1, warning_s, eee_getApplicationId("cm"), 0, 0, 0,
                      ("CM: Couldn't get fan #%u tach speed\n", fan));
                continue;
            }

            /* Check for a fan alarm or fan tach speed out of range condition
             */
            if ((alarm_val & FAN_ALARM_ENABLE_GPIO1) ||
                ((tach_val < min_rev) || (tach_val > max_rev))) {
                /* When fan is now faulty but was previously okay, send
                 * event indicating that fan has failed.
                 */
                if (fan_err_count[module][fan] == 0) {
                    cmSendFanEvent(module, fan, EVENT_STATE_DOWN,
                                   alarm_val, tach_val);
                } else if (!(fan_err_count[module][fan] % FAN_ERR_COUNT)) {
                    E_LOG(class_1, warning_s,
                          eee_getApplicationId("cm"), 0, 0, 0,
                          ("CM: Problem detected on fan module #%u fan #%u, "
                           "tach=%u rps, alarm=%02x (%s)\n",
                           module, fan, tach_val, alarm_val,
                           fan_alarm_msg(alarm_val)));
                }
                fan_err_count[module][fan]++;
                module_err = TRUE;
            } else {
                /* When fan is now ok but was previously in error, send
                 * event indicating that fan has recovered.
                 */
                if (fan_err_count[module][fan] != 0) {
                    cmSendFanEvent(module, fan, EVENT_STATE_UP,
                                   alarm_val, tach_val);
                    fan_err_count[module][fan] = 0;
                }
            }
        }

        /* Update Fan Module error count
         */
        if (module_err) {
            mod_err_count[module]++;
        } else {
            mod_err_count[module] = 0;
        }

        /* Update Fan Module LEDs
         */
        if (mm_readAddr(&val, sizeof(val), CM_BMFPGA_LED) == NFX_OK) {
            stat_val = (val >> (10 + (module * 2))) & 0x3;
            if (mod_err_count[module] != 0) {
                if (stat_val != CM_LED_RED) {
                    val &= ~(0x03 << (10 + (module * 2)));
                    val |= (CM_LED_RED << (10 + (module * 2)));
                    mm_writeAddr(&val, sizeof(val), CM_BMFPGA_LED);
                }
            } else {
                if (stat_val != CM_LED_OFF) {
                    val &= ~(0x03 << (10 + (module * 2)));
                    mm_writeAddr(&val, sizeof(val), CM_BMFPGA_LED);
                }
            }
        }
    }

    /* Release mezzanine I2C bus to allow accesses via other blade.
     */
    cm_unlock_mezz_i2c_bus();

    return NFX_OK;
}

/*++

Routine Description:
    This routine sets the GPP register System Up bit which updates the
    Front Panel System LED turning it from flashing green to solid green
    during initialization.

Arguments:
    none

Return Value:
    none

--*/
static void
cm_set_SYSUP(void)
{
    uint32 val, new_val;

    if (mm_readAddr(&val, sizeof(val), CM_BMFPGA_GPP) == NFX_OK) {
        new_val = val | CM_SYSTEM_UP;
        new_val &= ~(CM_BOOT_ENABLE | CM_CRASH_ENABLE);
        if (val != new_val) {
            mm_writeAddr(&new_val, sizeof(new_val), CM_BMFPGA_GPP);
        }
    }
}

/*++

Routine Description:
    This routine sets the GPP register Crash Enable bit which updates
    the Front Panel System LED turning it from solid green to red.

Arguments:
    none

Return Value:
    none

--*/
static void
cm_set_SYSFLT(void)
{
    uint32 val, new_val;

    if (mm_readAddr(&val, sizeof(val), CM_BMFPGA_GPP) == NFX_OK) {
        new_val = val | CM_CRASH_ENABLE;
        new_val &= ~(CM_BOOT_ENABLE | CM_SYSTEM_UP);
        if (val != new_val) {
            mm_writeAddr(&new_val, sizeof(new_val), CM_BMFPGA_GPP);
        }
    }
}

/*++

Routine Description:
    This routine sets the specified compact flash LED.

Arguments:
    led - CM_CF_DEV1 or CM_CF_DEV2
    state - CM_LED_OFF, CM_LED_RED, CM_LED_GREEN, CM_LED_AMBER

Return Value:
    none

--*/
void
cm_set_CFLED(uint32 led, uint32 state)
{
    uint32 val;

    if (mm_readAddr(&val, sizeof(val), CM_BMFPGA_LED) == NFX_OK) {
        if (led == CM_CF_DEV1) {
            val &= ~0x03;
            val |= (state & 0x3);
        } else if (led == CM_CF_DEV2) {
            val &= ~0x0c;
            val |= ((state & 0x3) << 2);
        }
        mm_writeAddr(&val, sizeof(val), CM_BMFPGA_LED);
    }
}

/*++

Routine Description:
    This is the periodic polling routine which is called at fixed time
    intervals to check on system status and etc.  Message polling is done 
    elsewhere.

Arguments:
    none

Return Value:
    none

--*/
static void
cm_poll(void)
{
    uint32 val;

    if (mm_readAddr(&val, sizeof(val), CM_BMFPGA_STATUS) != NFX_OK) {
        E_LOG(class_1, warning_s, eee_getApplicationId("cm"), 0, 0, 0,
             ("CM: Couldn't get BMFPGA board status\n"));
        return;
    }

    /* 
     * check power supply status
     * check fan status
     * check temperature alarm threshold
     */
    if (cm_checkPowerStatus(val) != NFX_OK) {
        E_LOG(class_1, warning_s, eee_getApplicationId("cm"), 0, 0, 0,
             ("CM: Couldn't check power supply status\n"));
    }
    if (cm_checkFanStatus(val) != NFX_OK) {
        E_LOG(class_1, warning_s, eee_getApplicationId("cm"), 0, 0, 0,
             ("CM: Couldn't check fan status\n"));
    }
    if (cm_checkTempStatus() != NFX_OK) {
        E_LOG(class_1, warning_s, eee_getApplicationId("cm"), 0, 0, 0,
             ("CM: Couldn't check system temperature\n"));
    }
}

/*++

Routine Description:
    This routine checks the CPU status of all modules and sends UP/DOWN
    events to the event manager. If any cpu is not in runtime state,
    the crash_enabled bit is set which causes hardware to change the
    Front Panel System LED from green to red.

Arguments:
    none

Return Value:
    none

--*/
static void
cm_cpuPoll(void)
{
    uint32 val, state, cpu_id, fault = 0, wait_crash_dump = 0;
    cm_cpu_t *cpu;
    uint64 mbval;
    int rc;

    rc = cm_mbox_read_reg(&mbval);
    if (rc != NFX_OK) {
        CM_DBG_PRINTF(("Couldn't check CPU status."));
        return;
    }

    for (cpu_id = 0; cpu_id <= NFX_NFP_FP_LAST; cpu_id++) {

        cpu = &cm.slot[NFX_SLOTTYPE_NFP].cpu[cpu_id];

        /* Decode cpu state from mailbox register value
         */
        if (((mbval >> 32) & (1 << (CM_NUM_CPUS + 2)))) {
            state = CM_SLICE_STATE_RESET_WAIT;
        } else if (((mbval >> 32) & (1 << (CM_NUM_CPUS + 1)))) {
            state = CM_SLICE_STATE_CRASH_DONE;
        } else if (((mbval >> 32) & (1 << CM_NUM_CPUS))) {
            state = CM_SLICE_STATE_RUNTIME_CRASHED;
        } else if (((mbval >> 32) & (1 << cpu_id))) {
            state = CM_SLICE_STATE_RUNTIME;
        } else {
            state = CM_SLICE_STATE_NOT_PRESENT;
        }

        if ((cpu->state != CM_SLICE_STATE_RUNTIME) &&
            (state == CM_SLICE_STATE_RUNTIME)) {
            if (cpu->state != CM_SLICE_STATE_RUNTIME_REBOOT) {
                if (cpu_id < NFX_NFP_FP1) {
                    CM_DBG_PRINTF(("TXRX Module, CPU %d UP\n", cpu_id));
                } else {
                    CM_DBG_PRINTF(("FP Module, CPU %d UP\n",
                                  cpu_id - NFX_NFP_FP1));
                }
                cmSendCpuEvent_cg(NFX_SLOTTYPE_NFP, cpu_id, EVENT_STATE_UP);
                cpu->downtime = 0;
            }
        } else { 
            if ((cpu->state == CM_SLICE_STATE_RUNTIME)
                && (state != CM_SLICE_STATE_RUNTIME)) {
                if (cpu_id < NFX_NFP_FP1) {
                    CM_DBG_PRINTF(("TXRX Module, CPU %d DOWN\n", cpu_id));
                } else {
                    CM_DBG_PRINTF(("FP Module, CPU %d DOWN\n",
                                  cpu_id - NFX_NFP_FP1));
                }
                cmSendCpuEvent_cg(NFX_SLOTTYPE_NFP, cpu_id, EVENT_STATE_DOWN);
            }
        }

        cpu->state = state;
        if (cpu->state != CM_SLICE_STATE_RUNTIME) {
            fault = 1;
            cpu->downtime += CM_SHORT_POLL_TIME_SECS;
            if ((cpu->state == CM_SLICE_STATE_RESET_WAIT) || 
                (cpu->state == CM_SLICE_STATE_CRASH_DONE) ||
                (cpu->downtime >= CM_DOWNTIME_MAX_MINUTES * SEC_PER_MIN)) {

                if (cmGetAutoreboot_flag()) {
                    char *pCpuList;

                    /* When this cpu is listed, skip rebooting the system
                     * because of it.
                     */
                    if ((cpu->state == CM_SLICE_STATE_NOT_PRESENT) &&
                        ((pCpuList = getenv("NO_CPU_LIST")) != NULL) &&
                        (strstr(pCpuList, cpuList[cpu_id]) != NULL)) {

                        E_LOG(class_1, info_s, eee_getApplicationId("cm"),
                              0,0,0, ("%s is not present.", cpuList[cpu_id]));
                        cpu->downtime = 0;
                        continue;
                    }

                    cmResetMacs();
                    cm_delay_sec_bc(2);
                    utils_reboot_system("Node going down for reboot! "
						"(chassisd: because FP/TXRX module is down).", "cm", 0);
                }
            }
            else 
            if (cpu->state == CM_SLICE_STATE_RUNTIME_CRASHED) {
                wait_crash_dump = 1;
            }
        }
    }

    if (fault > 0) {
        cm_set_SYSFLT();
    }
}

/*++

Routine Description:
    This is the initialization routine for the chassis manager called once
    during startup.

Arguments:
    none

Return Value:
    none

--*/
static void
chassisd_init(void)
{
    uint32   module, cpu, cf, i;
    uchar8 ver[CM_VER_STRING_LEN];
    seepPCB_t seep;
    boolean valid;
    int rc;

    for (module = 0; module < CM_NUM_SLOTS; module++) {
        cm.slot[module].present = 1;
        cm.slot[module].temp_valid = 0;
        cm.slot[module].power_valid = 0;
        cm.slot[module].control = 0xFF;
        cm.slot[module].downtime = 0;
        for (cpu = 0; cpu < CM_NUM_CPUS; cpu++) {
            cm.slot[module].cpu[cpu].downtime = 0;
            cm.slot[module].cpu[cpu].outtime = 0;
        }
    }
    cm.slot[NFX_SLOTTYPE_SSC].type = CM_CARD_TYPE_SSC;
    cm.slot[NFX_SLOTTYPE_SSC].num_cpus = 1;
    cm.slot[NFX_SLOTTYPE_SSC].cpu[0].state = CM_SLICE_STATE_RUNTIME;
    cm.slot[NFX_SLOTTYPE_NFP].type = CM_CARD_TYPE_BOBCAT;
    cm.slot[NFX_SLOTTYPE_NFP].num_cpus = NFX_NFP_NUM_CPUS;
    cmGetVersionSSC(&ver[0]);
    strcpy(cm.slot[NFX_SLOTTYPE_SSC].cpu[0].prom_ver, ver); 
    cmGetVersionBMFPGA(&ver[0]);
    strcpy(cm.slot[NFX_SLOTTYPE_SSC].bmfpga_ver, ver);
    cmGetBootDevInfo_bc(&ver[0], &cf);
    strcpy(cm.slot[NFX_SLOTTYPE_SSC].boot_dev, ver);

    cmSetWatchDog(CM_WATCHDOG_TIMEOUT);

    rc = cm_read_seep_bc((uchar8 *)&seep, 0, sizeof(seepPCB_t));
    if (rc != NFX_OK) {
        E_LOG(class_1, err_s, eee_getApplicationId("cm"), 0, 0, 0,
              ("CM: Can't read PCB SEEP\n"));
        valid = FALSE; 
    } else if (strcmp(seep.signature, SEEP_SIG_PCB) != 0) {
        for (i = 0; i < sizeof(seep.signature)-1; i++) {
            if (isascii(seep.signature[i]) == 0) {
                seep.signature[i] = '?';
            }
        }
        seep.signature[sizeof(seep.signature)-1] = '\0';
        E_LOG(class_1, err_s, eee_getApplicationId("cm"), 0, 0, 0,
              ("CM: PCB SEEP signature error (exp: %s got: %s)\n",
               SEEP_SIG_PCB, seep.signature));
        valid = FALSE; 
    } else {
        valid = TRUE;
    }
    if (valid) {
        sprintf(cm.slot[NFX_SLOTTYPE_SSC].chassis_macAddr, 
               "MAC addr: %c%c:%c%c:%c%c:%c%c:%c%c:%c%c", 
               seep.mac_addr[0], seep.mac_addr[1],
               seep.mac_addr[2], seep.mac_addr[3],
               seep.mac_addr[4], seep.mac_addr[5],
               seep.mac_addr[6], seep.mac_addr[7],
               seep.mac_addr[8], seep.mac_addr[9],
               seep.mac_addr[10], seep.mac_addr[11]);
        sprintf(cm.slot[NFX_SLOTTYPE_SSC].hw_ver,
               "Board revision: %s", seep.revision);
        sprintf(cm.slot[NFX_SLOTTYPE_SSC].hw_serialNo, 
               "Board serial number: %s", seep.serial_num);
    } else {
        strcpy(cm.slot[NFX_SLOTTYPE_SSC].chassis_macAddr, 
               "MAC addr: unknown");
        strcpy(cm.slot[NFX_SLOTTYPE_SSC].hw_ver, 
               "Board revision: unknown");
        strcpy(cm.slot[NFX_SLOTTYPE_SSC].hw_serialNo, 
               "Board serial number: unknown");
    }

    cmSetWatchDog(CM_WATCHDOG_TIMEOUT);

    if (cm_read_chassis_seep((uchar8 *)&chsSeep, 0,
                             sizeof(seepChassis_t)) != NFX_OK) {
        E_LOG(class_1, err_s, eee_getApplicationId("cm"), 0, 0, 0,
              ("CM: Can't read Chassis SEEP\n"));
        bzero(&chsSeep, sizeof(seepChassis_t));
        valid = FALSE; 
    } else if (strcmp(chsSeep.signature, SEEP_SIG_CHASSIS) != 0) {
        for (i = 0; i < sizeof(chsSeep.signature)-1; i++) {
            if (isascii(chsSeep.signature[i]) == 0) {
                chsSeep.signature[i] = '?';
            }
        }
        chsSeep.signature[sizeof(chsSeep.signature)-1] = '\0';
        E_LOG(class_1, err_s, eee_getApplicationId("cm"), 0, 0, 0,
              ("CM: Chassis SEEP signature error (exp: %s got: %s)\n",
               SEEP_SIG_CHASSIS, chsSeep.signature));
        bzero(&chsSeep, sizeof(seepChassis_t));
        valid = FALSE; 
    } else {
        valid = TRUE; 
    }
    if (valid) {
        sprintf(cm.slot[NFX_SLOTTYPE_SSC].chassis_serialNo, 
               "System serial number: %s", chsSeep.serial_num);
        sprintf(cm.slot[NFX_SLOTTYPE_SSC].chassis_hw, 
               "System revision: %s", chsSeep.revision);

        /* When MAC address is set in Chassis SEEP, it overrides
         * the value in PCB SEEP.
         */
        if (strlen(chsSeep.mac_addr) == sizeof(chsSeep.mac_addr) - 1) {
            sprintf(cm.slot[NFX_SLOTTYPE_SSC].chassis_macAddr, 
                   "MAC addr: %c%c:%c%c:%c%c:%c%c:%c%c:%c%c", 
                   chsSeep.mac_addr[0], chsSeep.mac_addr[1],
                   chsSeep.mac_addr[2], chsSeep.mac_addr[3],
                   chsSeep.mac_addr[4], chsSeep.mac_addr[5],
                   chsSeep.mac_addr[6], chsSeep.mac_addr[7],
                   chsSeep.mac_addr[8], chsSeep.mac_addr[9],
                   chsSeep.mac_addr[10], chsSeep.mac_addr[11]);
        }
     } else {
        strcpy(cm.slot[NFX_SLOTTYPE_SSC].chassis_serialNo, 
               "System serial number: unknown");
        strcpy(cm.slot[NFX_SLOTTYPE_SSC].chassis_hw, 
               "System revision: unknown");
    }

    cm_update_CFLED();
    cm_set_SYSUP();
}

/*++

Routine Description:
    This routine initializes the mmap interface and i2c semaphore. Then it
    loops forever monitoring the power supplies, fans, temperature and FP
    processor runtime status.

Arguments:
    none

Return Value:
    none
 
--*/
void chassisd_run(void)
{
    struct timeval tv, pv_long, pv_short;
    int rc;

    rc = cmSetWatchDog(CM_WATCHDOG_TIMEOUT);
    if (rc != 0) {
        E_LOG(class_1, err_s, eee_getApplicationId("cm"), 0, 0, 0,
             ("CM: Couldn't set watchdog timer.\n"));
    }
    if (mm_init() != 0) {
        E_LOG(class_1, err_s, eee_getApplicationId("cm"), 0, 0, 0,
              ("CM: Couldn't initialize mmap interface.\n"));
        exit(1);
    }
    if (cm_init_i2c_access() != NFX_OK) {
        E_LOG(class_1, err_s, eee_getApplicationId("cm"), 0, 0, 0,
              ("CM: Couldn't initialize i2c semaphore.\n"));
    }

    chassisd_init();

    pv_long.tv_usec = 0;
    pv_long.tv_sec = 0;
    pv_short.tv_usec = 0;
    pv_short.tv_sec = 0;

    while (1) {
        chassisd_rcvMsg();
        cmSetWatchDog(CM_WATCHDOG_TIMEOUT);

        /* if certain time interval has elapsed then poll for status
         */
        if (gettimeofday(&tv, NULL) == 0) {
            if (pv_long.tv_sec + CM_LONG_POLL_TIME_SECS < tv.tv_sec) {
                cm_poll();
                pv_long = tv;
            }
            if (pv_short.tv_sec + CM_SHORT_POLL_TIME_SECS < tv.tv_sec) {
                cm_cpuPoll();
                pv_short = tv;
            }

            /* resync to handle the case if time is set backwards
             */
            if (pv_short.tv_sec > tv.tv_sec) {
                pv_short = tv;
                pv_long = tv;
            }
        }
    }
}
#endif  /* SSC */

