/*
 *-----------------------------------------------------------------
 *
 * Name:	cm.h
 *
 * Copyright (C) 2002,2008 Onstor Inc.
 *
 * Date Created:	4/15/02
 *
 *-----------------------------------------------------------------
 */

#ifndef _CM_H
#define _CM_H

#define CM_DEBUG(x)		if (cm_debug) printf x
#define CM_DBG_PRINTF(x) printf x

#define CM_TEMP_CRITICAL	100
#define CM_TEMP_HIGH		85
#define CM_TEMP_NORMAL		50

/* Maximum card down time in minutes before requiring system reboot
 */
#define CM_DOWNTIME_MAX_MINUTES    20 

/* Threshold in minutes to determine if card is down indefinitely
 */
#define CM_DOWNTIME_THRESH_MINUTES  3

#if defined(COUGAR)
/* Cougar watchdog timer interval in seconds
 */
#define CM_WATCHDOG_TIMEOUT  5
/* bits to reset network MAC devices
 */
#define CM_GIGE_RESET_BITS CM_GIGE_SERDES_RST

#elif defined(BOBCAT)
/* Bobcat watchdog timer interval in seconds
 */
#define CM_WATCHDOG_TIMEOUT  16
/* bits to reset network MAC devices
 */
#define CM_GIGE_RESET_BITS (CM_LSI_MAC_RESET | CM_SERDES_RESET)
#else
/* Cheetah watchdog timer interval in seconds
 */
#define CM_WATCHDOG_TIMEOUT  30
#endif

/* polling time in seconds
 */
#define CM_SHORT_POLL_TIME_SECS 5  
#define CM_LONG_POLL_TIME_SECS 15  

/* number of errors before we repeat the warning
*/
#define PS_ERR_COUNT	240
#define FAN_ERR_COUNT	40
#define TEMP_ERR_COUNT	40

/* intake air temperature = measured temperature - 15 Celsius
*/
# define INTAKE_TEMP(temp)	((temp)-15)

/*
 * The cm_t structure is compiled and maintained by the chassisd daemon on
 * the SSC.
 *
 * Version info comes from direct MGMT plane messages from the other cards
 * in the system.  The rest of the info comes from communication with the
 * Toshiba CPU, which monitors BMFPGA info changes.
 */
typedef struct cm_cpu_struct {
	uchar8		prom_ver[CM_VER_STRING_LEN];
	uchar8		rt_ver[CM_VER_STRING_LEN];
	uint32		state;
	uint32		downtime;
    uint32      outtime;
} cm_cpu_t;

typedef struct cm_slot_struct {
	uint32		present;
	uint32		type;
	uint32		power_valid;
	uint32		power;
	uint32		temp_valid;
	uint32		temp;
	uint32		state;
	uchar8		bmfpga_ver[CM_VER_STRING_LEN];
	uchar8		luc_ver[CM_VER_STRING_LEN];
	uchar8		lmux_ver[CM_VER_STRING_LEN];
	uchar8		hw_ver[CM_VER_STRING_LEN];
	uchar8		hw_serialNo[CM_VER_STRING_LEN];
	uchar8		chassis_serialNo[CM_VER_STRING_LEN];
	uchar8		chassis_hw[CM_VER_STRING_LEN];
	uchar8		chassis_macAddr[CM_VER_STRING_LEN];
	uchar8		boot_dev[CM_VER_STRING_LEN];
	uint32		num_cpus;
	cm_cpu_t	cpu[CM_NUM_CPUS];
	uint32		control;			/* shadow copy of write-only reg */
	uint32		downtime;
} cm_slot_t;

typedef struct cm_struct {
	cm_slot_t	slot[CM_NUM_SLOTS];
} cm_t;

extern cm_t		cm;
extern uint32	cm_debug;

#define CM_FC1_GPIO_READ_STATE_MASK		0x00007800
#define CM_FC2_GPIO_READ_STATE_MASK		0x00078000
#define CM_FC3_GPIO_READ_STATE_MASK		0x00780000
#define CM_FC1_GPIO_READ_STATE_SHIFT		11
#define CM_FC2_GPIO_READ_STATE_SHIFT		15
#define CM_FC3_GPIO_READ_STATE_SHIFT		19
#define CM_FC_GPIO_WRITE_STATE_MASK		0x00078000
#define CM_FC_GPIO_WRITE_STATE_SHIFT		15

extern uint32	cm_ssc_port;
extern uint32	cm_txrx_port;
extern uint32	cm_fp_port;
extern uint32	cm_src_port;

extern int32	cm_appid;

extern uint32	cm_slot_status;

typedef struct {
	void	* msg;
	uint32	len;
	uint32	srcCpu;
	uint32	srcSlot;
	uchar8	srcApp[32];
} chassisd_msgDesc_t;

typedef struct {
    chassisd_msg_hdr_t  hdr;
	cm_t				cm;
} chassisdMsgDebugResp_t;

/*++

Routine Description:

    This routine disables the watchdog timer through the kernel.

Arguments:

    None

Return Value:

    0  - success
    -1 - failure

--*/

int cmDisableWatchDog(void);

/*++

Routine Description:

    This routine sets the watchdog timer through the kernel.

Arguments:

    sec - watchdog timer expiration in seconds

Return Value:

    0  - success
    -1 - failure

--*/

int cmSetWatchDog(uint32 sec);

int chassisd_msgInit(void);
#if defined(BOBCAT) || defined(COUGAR)
void chassisd_run(void);
#elif defined(CHEETAH)
int	cm_sendInfoReq(void);
#endif
int chassisd_rcvMsg(void);
int	cm_sendVersionReq(int slot, int cpu);
void cmSendNodeEvent(char *reason);

/*++

Routine Description:
    This routine sends a temperature event to the event manager

Arguments:
    temp  - temperature in celsius
    state - operational state (EVENT_STATE_UP, EVENT_STATE_DOWN)

Return Value:
    none

--*/
void
cmSendTemperatureEvent(uint32 temp, int state);

/*++

Routine Description:
    Reset the MAC device for FP ports.  This function is called
    after a crash to prevent flooding of pause frames.

Arguments:
    None

Return Value:
    None

--*/
void
cmResetMacs(void);

/*++

Routine Description:
    This routine checks the intake temperature status and sends alarms if
    necessary.  When temperature has risen above max threshold, an event
    message is sent as a notification of the error.  Otherwise, if the
    the temperature has remained above the max threshold for TEMP_ERR_COUNT
    times in a row or has gone higher, an ELOG message is periodically sent.
    When temperature subsides at least 3 degrees below the max threshold,
    an event message is sent as a notification of a recovery.

Arguments:
    none

Return Value:
    NFX_OK or NFX_ERR

--*/
int
cm_checkTempStatus(void);

/*++

Routine Description:
    This routine gets the SSC PROM version string.

Arguments:
    verStr - pointer to version string returned
             string will be null terminated, and there should be space for 
             at least CM_VER_STRING_LEN chars including null terminator

Return Value:
    None

--*/
void
cmGetVersionSSC(uchar8 *verStr);

/*++

Routine Description:
    This routine sets the specified compact flash LED.

Arguments:
    led - CM_CF_DEV1 or CM_CF_DEV2
    state - CM_LED_OFF, CM_LED_RED, CM_LED_GREEN, CM_LED_AMBER

Return Value:
    none

--*/
void
cm_set_CFLED(uint32 led, uint32 state);

/*++

Routine Description:
    This routine checks the boot device and sets the compact flash LED that
    was used (and still mounted) to amber.  The other compact flash LED of
    the other device is set to green.  When a network boot occurs, the LED
    of both compact flashes are set to green.

Arguments
    none

Return Value
    none

--*/
void
cm_update_CFLED(void);

#endif	/* _CM_H */

